"""
Rdio OAuth1 and OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/rdio.html
"""

from .oauth import BaseOAuth1, BaseOAuth2, OAuthAuth

RDIO_API = "https://www.rdio.com/api/1/"


class BaseRdio(OAuthAuth):
    ID_KEY = "key"

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            fullname=response["displayName"],
            first_name=response["firstName"],
            last_name=response["lastName"],
        )
        return {
            "username": response["username"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }


class RdioOAuth1(BaseRdio, BaseOAuth1):
    """Rdio OAuth authentication backend"""

    name = "rdio-oauth1"
    REQUEST_TOKEN_URL = "http://api.rdio.com/oauth/request_token"
    AUTHORIZATION_URL = "https://www.rdio.com/oauth/authorize"
    ACCESS_TOKEN_URL = "http://api.rdio.com/oauth/access_token"
    EXTRA_DATA = [
        ("key", "rdio_id"),
        ("icon", "rdio_icon_url"),
        ("url", "rdio_profile_url"),
        ("username", "rdio_username"),
        ("streamRegion", "rdio_stream_region"),
    ]

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        params = {
            "method": "currentUser",
            "extras": "username,displayName,streamRegion",
        }
        request = self.oauth_request(access_token, RDIO_API, params, method="POST")
        # TODO: I don't think to_postdata exists.
        return self.get_json(request.url, method="POST", data=request.to_postdata())[  # type: ignore[reportAttributeAccessIssue]
            "result"
        ]


class RdioOAuth2(BaseRdio, BaseOAuth2):
    name = "rdio-oauth2"
    AUTHORIZATION_URL = "https://www.rdio.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://www.rdio.com/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [
        ("key", "rdio_id"),
        ("icon", "rdio_icon_url"),
        ("url", "rdio_profile_url"),
        ("username", "rdio_username"),
        ("streamRegion", "rdio_stream_region"),
        ("refresh_token", "refresh_token", True),
        ("token_type", "token_type", True),
    ]

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            RDIO_API,
            method="POST",
            data={
                "method": "currentUser",
                "extras": "username,displayName,streamRegion",
                "access_token": access_token,
            },
        )["result"]



import base64

from .oauth import BaseOAuth2


class ZoomOAuth2(BaseOAuth2):
    """
    Zoom OAuth2 authentication backend
    Doc Reference: https://marketplace.zoom.us/docs/guides/auth/oauth
    """

    name = "zoom-oauth2"
    AUTHORIZATION_URL = "https://zoom.us/oauth/authorize"
    ACCESS_TOKEN_URL = "https://zoom.us/oauth/token"
    USER_DETAILS_URL = "https://api.zoom.us/v2/users/me"
    DEFAULT_SCOPE = ["user:read"]
    ACCESS_TOKEN_METHOD = "POST"
    REFRESH_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    EXTRA_DATA = [("expires_in", "expires")]

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            self.USER_DETAILS_URL,
            headers={"Authorization": f"Bearer {access_token}"},
        )

    def get_user_details(self, response):
        username = response.get("id", "")
        first_name = response.get("first_name", "")
        last_name = response.get("last_name", "")
        email = response.get("email", "")
        fullname = ""
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def auth_complete_params(self, state=None):
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "redirect_uri": self.get_redirect_uri(state),
        }

    def auth_headers(self):
        return {
            "Authorization": b"Basic "
            + base64.urlsafe_b64encode(
                "{}:{}".format(*self.get_key_and_secret()).encode()
            )
        }

    def refresh_token_params(self, token, *args, **kwargs):
        return {"refresh_token": token, "grant_type": "refresh_token"}



import json

from httpretty import HTTPretty

from ...exceptions import AuthFailed
from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class GithubOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.github.GithubOAuth2"
    user_data_url = "https://api.github.com/user"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "expires_in": 28800,
            "refresh_token": "foobar-refresh-token",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 28800,
            "refresh_token": "foobar-new-refresh-token",
            "refresh_token_expires_in": 15897600,
            "scope": "",
        }
    )
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def do_login(self):
        user = super().do_login()
        social = user.social[0]

        self.assertIsNotNone(social.extra_data["expires"])
        self.assertIsNotNone(social.extra_data["refresh_token"])

        return user

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")


class GithubOAuth2NoEmailTest(GithubOAuth2Test):
    user_data_body = json.dumps(
        {
            "login": "foobar",
            "id": 1,
            "avatar_url": "https://github.com/images/error/foobar_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar",
            "name": "monalisa foobar",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_login(self):
        url = "https://api.github.com/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps(["foo@bar.com"]),
            content_type="application/json",
        )
        self.do_login()

    def test_login_next_format(self):
        url = "https://api.github.com/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps([{"email": "foo@bar.com"}]),
            content_type="application/json",
        )
        self.do_login()

    def test_partial_pipeline(self):
        url = "https://api.github.com/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps([{"email": "foo@bar.com"}]),
            content_type="application/json",
        )
        self.do_partial_pipeline()

    def test_refresh_token(self):
        url = "https://api.github.com/user/emails"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=200,
            body=json.dumps([{"email": "foo@bar.com"}]),
            content_type="application/json",
        )
        self.do_refresh_token()


class GithubOrganizationOAuth2Test(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubOrganizationOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/orgs/foobar/members/foobar"
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_partial_pipeline()

    def test_refresh_token(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        self.do_refresh_token()


class GithubOrganizationOAuth2FailTest(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubOrganizationOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/orgs/foobar/members/foobar"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()

    def test_refresh_token(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_ORG_NAME": "foobar"})
        with self.assertRaises(AuthFailed):
            self.do_refresh_token()


class GithubTeamOAuth2Test(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/teams/123/members/foobar"
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body="")
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_partial_pipeline()

    def test_refresh_token(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        self.do_refresh_token()


class GithubTeamOAuth2FailTest(GithubOAuth2Test):
    backend_path = "social_core.backends.github.GithubTeamOAuth2"

    def auth_handlers(self, start_url):
        url = "https://api.github.com/teams/123/members/foobar"
        HTTPretty.register_uri(
            HTTPretty.GET,
            url,
            status=404,
            body='{"message": "Not Found"}',
            content_type="application/json",
        )
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()

    def test_refresh_token(self):
        self.strategy.set_settings({"SOCIAL_AUTH_GITHUB_TEAM_ID": "123"})
        with self.assertRaises(AuthFailed):
            self.do_refresh_token()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class StravaOAuthTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.strava.StravaOAuth"
    user_data_url = "https://www.strava.com/api/v3/athlete"
    expected_username = "marianne_v"
    access_token_body = json.dumps(
        {
            "token_type": "Bearer",
            "expires_at": 1572805000,
            "expires_in": 227615,
            "refresh_token": "f51defab4632d27255dd0d106504dfd7568fd1df6",
            "access_token": "83ebeabdec09f6670863766f792ead24d61fe3f9",
            "athlete": {
                "id": 1234567890987654321,
                "username": "marianne_v",
                "resource_state": 2,
                "firstname": "Marianne",
                "lastname": "V.",
                "city": "Francisco",
                "state": "California",
                "country": "United States",
                "sex": "F",
                "premium": "true",
                "summit": "true",
                "created_at": "2017-11-14T02:30:05Z",
                "updated_at": "2018-02-06T19:32:20Z",
                "badge_type_id": 4,
                "profile_medium": "https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg",
                "profile": "https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg",
                "friend": "null",
                "follower": "null",
            },
        }
    )
    user_data_body = json.dumps(
        {
            "id": 1234567890987654321,
            "username": "marianne_v",
            "resource_state": 3,
            "firstname": "Marianne",
            "lastname": "V.",
            "city": "San Francisco",
            "state": "CA",
            "country": "US",
            "sex": "F",
            "premium": "true",
            "created_at": "2017-11-14T02:30:05Z",
            "updated_at": "2018-02-06T19:32:20Z",
            "badge_type_id": 4,
            "profile_medium": "https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg",
            "profile": "https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg",
            "friend": "null",
            "follower": "null",
            "follower_count": 5,
            "friend_count": 5,
            "mutual_friend_count": 0,
            "athlete_type": 1,
            "date_preference": "%m/%d/%Y",
            "measurement_preference": "feet",
            "clubs": [],
            "ftp": "null",
            "weight": 0,
            "bikes": [],
            "shoes": [],
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from social_core.backends.oauth import BaseOAuth2


class MusicBrainzOAuth2(BaseOAuth2):
    """MusicBrainz OAuth authentication backend"""

    name = "musicbrainz"
    AUTHORIZATION_URL = "https://musicbrainz.org/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://musicbrainz.org/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    ID_KEY = "metabrainz_user_id"
    DEFAULT_SCOPE = ["profile", "email"]
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("metabrainz_user_id", "id"),
        ("expires_in", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details from MusicBrainz account"""
        return {
            "username": response.get("sub"),
            "email": response.get("email") or "",
            "first_name": response.get("sub"),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://musicbrainz.org/oauth2/userinfo",
            params={"access_token": access_token},
        )



"""
Beats backend, docs at:
    https://developer.beatsmusic.com/docs
"""

import base64

from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class BeatsOAuth2(BaseOAuth2):
    name = "beats"
    SCOPE_SEPARATOR = " "
    ID_KEY = "user_context"
    AUTHORIZATION_URL = "https://partner.api.beatsmusic.com/v1/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://partner.api.beatsmusic.com/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response["result"][BeatsOAuth2.ID_KEY]

    def auth_headers(self):
        return {
            "Authorization": "Basic {}".format(
                base64.urlsafe_b64encode(
                    "{}:{}".format(*self.get_key_and_secret()).encode()
                )
            )
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        # mashery wraps in jsonrpc
        if response.get("jsonrpc", None):
            response = response.get("result", None)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )

    def get_user_details(self, response):
        """Return user details from Beats account"""
        response = response["result"]
        fullname, first_name, last_name = self.get_user_names(
            response.get("display_name")
        )
        return {
            "username": response.get("id"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://partner.api.beatsmusic.com/v1/api/me",
            headers={"Authorization": f"Bearer {access_token}"},
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class PatreonOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.patreon.PatreonOAuth2"
    user_data_url = (
        "https://www.patreon.com/api/oauth2/v2/identity?"
        + "fields%5Buser%5D=about,created,email,first_name,full_name,"
        + "image_url,last_name,social_connections,thumb_url,url,vanity"
    )
    expected_username = "JohnInterwebs"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
        }
    )
    user_data_body = json.dumps(
        {
            "data": {
                "relationships": {
                    "pledges": {"data": [{"type": "pledge", "id": "123456"}]}
                },
                "attributes": {
                    "last_name": "Interwebs",
                    "is_suspended": False,
                    "has_password": True,
                    "full_name": "John Interwebs",
                    "is_nuked": False,
                    "first_name": "John",
                    "social_connections": {
                        "spotify": None,
                        "discord": None,
                        "twitter": None,
                        "youtube": None,
                        "facebook": None,
                        "deviantart": None,
                        "twitch": None,
                    },
                    "twitter": None,
                    "is_email_verified": True,
                    "facebook_id": None,
                    "email": "john@example.com",
                    "facebook": None,
                    "thumb_url": "https://c8.patreon.com/100/123456",
                    "vanity": None,
                    "about": None,
                    "is_deleted": False,
                    "created": "2017-05-05T05:16:34+00:00",
                    "url": "https://www.patreon.com/user?u=123456",
                    "gender": 0,
                    "youtube": None,
                    "discord_id": None,
                    "image_url": "https://c8.patreon.com/400/123456",
                    "twitch": None,
                },
                "type": "user",
                "id": "123456",
            }
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SlackOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.slack.SlackOAuth2"
    user_data_url = "https://slack.com/api/users.identity"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "ok": True,
            "user": {"email": "foobar@example.com", "name": "Foo Bar", "id": "123456"},
            "team": {"id": "456789"},
            "scope": "identity.basic,identity.email",
        }
    )
    expected_username = "foobar"

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class SlackOAuth2TestTeamName(SlackOAuth2Test):
    expected_username = "foobar@Square"
    user_data_body = json.dumps(
        {
            "ok": True,
            "user": {"email": "foobar@example.com", "name": "Foo Bar", "id": "123456"},
            "team": {
                "id": "456789",
                "name": "Square",
            },
            "scope": "identity.basic,identity.email,identity.team",
        }
    )


class SlackOAuth2TestUnicodeTeamName(SlackOAuth2Test):
    user_data_body = json.dumps(
        {
            "ok": True,
            "user": {"email": "foobar@example.com", "name": "Foo Bar", "id": "123456"},
            "team": {
                "id": "456789",
                "name": "Square \u221a team",
            },
            "scope": "identity.basic,identity.email,identity.team",
        }
    )

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_SLACK_USERNAME_WITH_TEAM": False})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_SLACK_USERNAME_WITH_TEAM": False})
        self.do_partial_pipeline()



"""
Orbi OAuth2 backend
"""

from .oauth import BaseOAuth2


class OrbiOAuth2(BaseOAuth2):
    """Orbi OAuth2 authentication backend"""

    name = "orbi"
    AUTHORIZATION_URL = "https://login.orbi.kr/oauth/authorize"
    ACCESS_TOKEN_URL = "https://login.orbi.kr/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"
    EXTRA_DATA = [
        ("imin", "imin"),
        ("nick", "nick"),
        ("photo", "photo"),
        ("sex", "sex"),
        ("birth", "birth"),
    ]

    def get_user_id(self, details, response):
        return response.get("id")

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("name", ""),
            response.get("first_name", ""),
            response.get("last_name", ""),
        )
        return {
            "username": response.get("username", response.get("name")),
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from orbi"""
        return self.get_json(
            "https://login.orbi.kr/oauth/user/get",
            params={"access_token": access_token},
        )



from .oauth import BaseOAuth2


class CleverOAuth2(BaseOAuth2):
    """
    Clever OAuth authentication backend.

    Docs: https://dev.clever.com/docs/classroom-with-oauth
    """

    name = "clever"
    AUTHORIZATION_URL = "https://clever.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://clever.com/oauth/tokens"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    STATE_PARAMETER = False
    SCOPE_SEPARATOR = " "

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response.get("data", {}).get("id")

    def get_user_type(self, data):
        return next(iter(data.get("data", {}).get("roles", {}).keys()))

    def get_user_details(self, response):
        """Return user details from Classlink account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("data", {}).get("name", {}).get("first", None),
            last_name=response.get("data", {}).get("name", {}).get("last", None),
        )
        email = response.get("data", {}).get("email")
        username = (
            response.get("data", {})
            .get("roles", {})
            .get(self.get_user_type(response), {})
            .get("credentials", {})
            .get("district_username", email.split("@", 1)[0])
        )
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "user_type": self.get_user_type(response),
        }

    def user_data(self, token, *args, **kwargs):
        """Loads user data from service"""
        identity_url = "https://api.clever.com/v3.0/me"
        user_details_url = "https://api.clever.com/v3.0/users"
        auth_header = {"Authorization": f"Bearer {token}"}
        try:
            response = self.get_json(identity_url, headers=auth_header)
            user_id = response.get("data", {}).get("id")
            user_details_url = f"https://api.clever.com/v3.0/users/{user_id}"
            return self.get_json(user_details_url, headers=auth_header)
        except ValueError:
            return None



"""
LINE Login OAuth2 backend, docs at:
    https://developers.line.me/en/docs/line-login/
"""

import json

import requests

from ..exceptions import AuthFailed
from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class LineOAuth2(BaseOAuth2):
    name = "line"
    AUTHORIZATION_URL = "https://access.line.me/oauth2/v2.1/authorize"
    ACCESS_TOKEN_URL = "https://api.line.me/oauth2/v2.1/token"
    BASE_API_URL = "https://api.line.me"
    USER_INFO_URL = BASE_API_URL + "/v2/profile"
    ACCESS_TOKEN_METHOD = "POST"
    STATE_PARAMETER = True
    DEFAULT_SCOPE = ["profile"]
    REDIRECT_STATE = True
    ID_KEY = "userId"
    EXTRA_DATA = [
        ("userId", "id"),
        ("picture_url", "picture_url"),
        ("status_message", "status_message"),
        ("expires_in", "expire"),
        ("refresh_token", "refresh_token"),
    ]

    def auth_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "response_type": self.RESPONSE_TYPE,
            "client_id": client_id,
            "redirect_uri": self.get_redirect_uri(),
            "state": self.get_or_create_state(),
            "scope": self.get_scope(),
        }

    def process_error(self, data):
        error_code = (
            data.get("errorCode") or data.get("statusCode") or data.get("error")
        )
        error_message = data.get("errorMessage") or data.get("error_description")
        if error_code is not None or error_message is not None:
            raise AuthFailed(self, error_message or error_code)

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)

        try:
            response = self.request_access_token(
                self.access_token_url(),
                method=self.ACCESS_TOKEN_METHOD,
                headers=self.auth_headers(),
                data=self.auth_complete_params(),
            )
            self.process_error(response)

            return self.do_auth(
                response["access_token"], response=response, *args, **kwargs
            )
        except requests.HTTPError as err:
            self.process_error(json.loads(err.response.content))

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get("displayName")
        )
        username = response.get("userId")
        picture_url = response.get("pictureUrl")
        status_message = response.get("statusMessage")
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "picture_url": picture_url,
            "status_message": status_message,
        }

    def get_user_id(self, details, response):
        """
        Return a unique ID for the current user, by default from
        server response.
        """
        return response.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        try:
            response = self.get_json(
                self.USER_INFO_URL, headers={"Authorization": f"Bearer {access_token}"}
            )
            self.process_error(response)
            return response
        except requests.HTTPError as err:
            self.process_error(err.response.json())



"""
Discogs OAuth1 backend, docs at:
    https://www.discogs.com/developers/
"""

from social_core.backends.oauth import BaseOAuth1


class DiscogsOAuth1(BaseOAuth1):
    """
    Implements the OAuth1 authentication mechanism for https://www.discogs.com
    """

    name = "discogs"

    OAUTH_TOKEN_PARAMETER_NAME = "oauth_token"

    AUTHORIZATION_URL = "https://www.discogs.com/oauth/authorize"
    REQUEST_TOKEN_URL = "https://api.discogs.com/oauth/request_token"
    ACCESS_TOKEN_URL = "https://api.discogs.com/oauth/access_token"

    def get_user_details(self, user_data):  # type: ignore[reportIncompatibleMethodOverride]
        return {
            "username": user_data["username"],
            "id": user_data["id"],
            "profile": user_data["profile"],
            "name": user_data["name"],
        }

    def user_data(self, access_token, *args, **kwargs):
        identity = self.get_json(
            "https://api.discogs.com/oauth/identity", auth=self.oauth_auth(access_token)
        )

        return self.get_json(
            identity["resource_url"], auth=self.oauth_auth(access_token)
        )



from urllib.parse import urlencode

from requests_oauthlib import OAuth1

from .oauth import BaseOAuth2


class NKOAuth2(BaseOAuth2):
    """NK OAuth authentication backend"""

    name = "nk"
    AUTHORIZATION_URL = "https://nk.pl/oauth2/login"
    ACCESS_TOKEN_URL = "https://nk.pl/oauth2/token"
    SCOPE_SEPARATOR = ","
    ACCESS_TOKEN_METHOD = "POST"
    SIGNATURE_TYPE_AUTH_HEADER = "AUTH_HEADER"
    EXTRA_DATA = [
        ("id", "id"),
    ]

    def get_user_details(self, response):
        """Return user details from NK account"""
        entry = response["entry"]
        return {
            "username": entry.get("displayName"),
            "email": entry["emails"][0]["value"],
            "first_name": entry.get("displayName").split(" ")[0],
            "id": entry.get("id"),
        }

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uri": self.get_redirect_uri(state),
            "scope": self.get_scope_argument(),
        }

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response."""
        return details.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = "http://opensocial.nk-net.pl/v09/social/rest/people/@me?" + urlencode(
            {
                "nk_token": access_token,
                "fields": "name,surname,avatar,localization,age,"
                + "gender,emails,birthdate",
            }
        )
        return self.get_json(url, auth=self.oauth_auth(access_token))

    def oauth_auth(
        self, token=None, oauth_verifier=None, signature_type=SIGNATURE_TYPE_AUTH_HEADER
    ):
        key, secret = self.get_key_and_secret()
        oauth_verifier = oauth_verifier or self.data.get("oauth_verifier")
        token = token or {}
        state = self.get_or_create_state()
        return OAuth1(
            key,
            secret,
            resource_owner_key=None,
            resource_owner_secret=None,
            callback_uri=self.get_redirect_uri(state),
            verifier=oauth_verifier,
            signature_type=signature_type,
        )



from .oauth import BaseOAuth2


class CILogonOAuth2(BaseOAuth2):
    """
    CI Logon Authentication Backend

    Docs: https://www.cilogon.org/oidc
    """

    name = "cilogon-oauth2"
    AUTHORIZATION_URL = "https://cilogon.org/authorize"
    ACCESS_TOKEN_URL = "https://cilogon.org/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    DEFAULT_SCOPE = ["openid", "email", "profile", "org.cilogon.userinfo"]
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = "+"

    def user_data(self, token, *args, **kwargs):
        """Loads user data from endpoint"""
        url = "https://cilogon.org/oauth2/userinfo"
        data = {"access_token": token}
        try:
            return self.get_json(url, method="POST", data=data)
        except ValueError:
            return None

    def get_user_id(self, details, response):
        """Return user unique id provided by service
        In this case it is a combination of the `sub`
        and `iss` respective values."""
        return response.get("sub", "") + " " + response.get("iss", "")

    def get_user_details(self, response):
        """Return user details from CI Logon service"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("given_name"), last_name=response.get("family_name")
        )
        return {
            "username": response.get("email"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



import unittest

from ...backends.github import GithubOAuth2
from ...backends.utils import get_backend, load_backends
from ...exceptions import MissingBackend
from ..models import TestStorage
from ..strategy import TestStrategy


class BaseBackendUtilsTest(unittest.TestCase):
    def setUp(self):
        self.strategy = TestStrategy(storage=TestStorage)

    def tearDown(self):
        self.strategy = None


class LoadBackendsTest(BaseBackendUtilsTest):
    def test_load_backends(self):
        loaded_backends = load_backends(
            (
                "social_core.backends.github.GithubOAuth2",
                "social_core.backends.facebook.FacebookOAuth2",
                "social_core.backends.flickr.FlickrOAuth",
            ),
            force_load=True,
        )
        keys = list(loaded_backends.keys())
        self.assertEqual(keys, ["github", "facebook", "flickr"])

        backends = ()
        loaded_backends = load_backends(backends, force_load=True)
        self.assertEqual(len(list(loaded_backends.keys())), 0)


class GetBackendTest(BaseBackendUtilsTest):
    def test_get_backend(self):
        backend = get_backend(
            (
                "social_core.backends.github.GithubOAuth2",
                "social_core.backends.facebook.FacebookOAuth2",
                "social_core.backends.flickr.FlickrOAuth",
            ),
            "github",
        )
        self.assertEqual(backend, GithubOAuth2)

    def test_get_missing_backend(self):
        with self.assertRaisesRegex(MissingBackend, 'Missing backend "foobar" entry'):
            get_backend(
                (
                    "social_core.backends.github.GithubOAuth2",
                    "social_core.backends.facebook.FacebookOAuth2",
                    "social_core.backends.flickr.FlickrOAuth",
                ),
                "foobar",
            )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ClefOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.clef.ClefOAuth2"
    user_data_url = "https://clef.io/api/v1/info"
    expected_username = "test"
    access_token_body = json.dumps({"access_token": "foobar"})
    user_data_body = json.dumps(
        {
            "info": {
                "id": "123456789",
                "first_name": "Test",
                "last_name": "User",
                "email": "test@example.com",
            }
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...exceptions import AuthForbidden
from .oauth import BaseAuthUrlTestMixin, OAuth1AuthUrlTestMixin, OAuth1Test, OAuth2Test


class BitbucketOAuthMixin:
    user_data_url = "https://api.bitbucket.org/2.0/user"
    expected_username = "foobar"
    bb_api_user_emails = "https://api.bitbucket.org/2.0/user/emails"

    user_data_body = json.dumps(
        {
            "created_on": "2012-03-29T18:07:38+00:00",
            "display_name": "Foo Bar",
            "links": {
                "avatar": {"href": "https://bitbucket.org/account/foobar/avatar/32/"},
                "followers": {
                    "href": "https://api.bitbucket.org/2.0/users/foobar/followers"
                },
                "following": {
                    "href": "https://api.bitbucket.org/2.0/users/foobar/following"
                },
                "hooks": {"href": "https://api.bitbucket.org/2.0/users/foobar/hooks"},
                "html": {"href": "https://bitbucket.org/foobar"},
                "repositories": {
                    "href": "https://api.bitbucket.org/2.0/repositories/foobar"
                },
                "self": {"href": "https://api.bitbucket.org/2.0/users/foobar"},
            },
            "location": "Fooville, Bar",
            "type": "user",
            "username": "foobar",
            "uuid": "{397621dc-0f78-329f-8d6d-727396248e3f}",
            "website": "http://foobar.com",
        }
    )

    emails_body = json.dumps(
        {
            "page": 1,
            "pagelen": 10,
            "size": 2,
            "values": [
                {
                    "email": "foo@bar.com",
                    "is_confirmed": True,
                    "is_primary": True,
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/user/emails/foo@bar.com"
                        }
                    },
                    "type": "email",
                },
                {
                    "email": "not@confirme.com",
                    "is_confirmed": False,
                    "is_primary": False,
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/user/emails/not@confirmed.com"
                        }
                    },
                    "type": "email",
                },
            ],
        }
    )


class BitbucketOAuth1Test(BitbucketOAuthMixin, OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.bitbucket.BitbucketOAuth"

    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )

    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})

    def test_login(self):
        HTTPretty.register_uri(
            HTTPretty.GET, self.bb_api_user_emails, status=200, body=self.emails_body
        )
        self.do_login()

    def test_partial_pipeline(self):
        HTTPretty.register_uri(
            HTTPretty.GET, self.bb_api_user_emails, status=200, body=self.emails_body
        )
        self.do_partial_pipeline()


class BitbucketOAuth1FailTest(BitbucketOAuth1Test):
    emails_body = json.dumps(
        {
            "page": 1,
            "pagelen": 10,
            "size": 1,
            "values": [
                {
                    "email": "foo@bar.com",
                    "is_confirmed": False,
                    "is_primary": True,
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/user/emails/foo@bar.com"
                        }
                    },
                    "type": "email",
                }
            ],
        }
    )

    def test_login(self):
        self.strategy.set_settings({"SOCIAL_AUTH_BITBUCKET_VERIFIED_EMAILS_ONLY": True})
        with self.assertRaises(AuthForbidden):
            super().test_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({"SOCIAL_AUTH_BITBUCKET_VERIFIED_EMAILS_ONLY": True})
        with self.assertRaises(AuthForbidden):
            super().test_partial_pipeline()


class BitbucketOAuth2Test(BitbucketOAuthMixin, OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.bitbucket.BitbucketOAuth2"

    access_token_body = json.dumps(
        {
            "access_token": "foobar_access",
            "scopes": "foo_scope",
            "expires_in": 3600,
            "refresh_token": "foobar_refresh",
            "token_type": "bearer",
        }
    )

    def test_login(self):
        HTTPretty.register_uri(
            HTTPretty.GET, self.bb_api_user_emails, status=200, body=self.emails_body
        )
        self.do_login()

    def test_partial_pipeline(self):
        HTTPretty.register_uri(
            HTTPretty.GET, self.bb_api_user_emails, status=200, body=self.emails_body
        )
        self.do_partial_pipeline()


class BitbucketOAuth2FailTest(BitbucketOAuth2Test):
    emails_body = json.dumps(
        {
            "page": 1,
            "pagelen": 10,
            "size": 1,
            "values": [
                {
                    "email": "foo@bar.com",
                    "is_confirmed": False,
                    "is_primary": True,
                    "links": {
                        "self": {
                            "href": "https://api.bitbucket.org/2.0/user/emails/foo@bar.com"
                        }
                    },
                    "type": "email",
                }
            ],
        }
    )

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_BITBUCKET_OAUTH2_VERIFIED_EMAILS_ONLY": True}
        )
        with self.assertRaises(AuthForbidden):
            super().test_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_BITBUCKET_OAUTH2_VERIFIED_EMAILS_ONLY": True}
        )
        with self.assertRaises(AuthForbidden):
            super().test_partial_pipeline()



import json

from ...backends.facebook import API_VERSION
from ...exceptions import AuthCanceled, AuthUnknownError
from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .test_open_id_connect import OpenIdConnectTestMixin


class FacebookOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.facebook.FacebookOAuth2"
    user_data_url = f"https://graph.facebook.com/v{API_VERSION}/me"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "username": "foobar",
            "first_name": "Foo",
            "last_name": "Bar",
            "verified": True,
            "name": "Foo Bar",
            "gender": "male",
            "updated_time": "2013-02-13T14:59:42+0000",
            "link": "http://www.facebook.com/foobar",
            "id": "110011001100010",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class FacebookOAuth2WrongUserDataTest(FacebookOAuth2Test):
    user_data_body = "null"

    def test_login(self):
        with self.assertRaises(AuthUnknownError):
            self.do_login()

    def test_partial_pipeline(self):
        with self.assertRaises(AuthUnknownError):
            self.do_partial_pipeline()


class FacebookOAuth2AuthCancelTest(FacebookOAuth2Test):
    access_token_status = 400
    access_token_body = json.dumps(
        {
            "error": {
                "message": "redirect_uri isn't an absolute URI. Check RFC 3986.",
                "code": 191,
                "type": "OAuthException",
                "fbtrace_id": "123Abc",
            }
        }
    )

    def test_login(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_login()
        self.assertIn("error", cm.exception.response.json())

    def test_partial_pipeline(self):
        with self.assertRaises(AuthCanceled) as cm:
            self.do_partial_pipeline()
        self.assertIn("error", cm.exception.response.json())


class FacebookLimitedLoginTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = "social_core.backends.facebook_limited.FacebookLimitedLogin"
    issuer = "https://facebook.com"
    openid_config_body = """
    {
      "issuer": "https://facebook.com",
      "authorization_endpoint": "https://facebook.com/dialog/oauth/",
      "jwks_uri": "https://facebook.com/.well-known/oauth/openid/jwks/",
      "response_types_supported": [
        "id_token",
        "token id_token"
      ],
      "subject_types_supported": "pairwise",
      "id_token_signing_alg_values_supported": [
        "RS256"
      ],
      "claims_supported": [
        "iss",
        "aud",
        "sub",
        "iat",
        "exp",
        "jti",
        "nonce",
        "at_hash",
        "name",
        "email",
        "picture",
        "user_friends",
        "user_birthday",
        "user_age_range"
      ]
    }
    """

    def test_invalid_nonce(self):
        # The nonce isn't generated server-side so the test isn't relevant here.
        pass



"""
Drip OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/drip.html
"""

from .oauth import BaseOAuth2


class DripOAuth(BaseOAuth2):
    name = "drip"
    AUTHORIZATION_URL = "https://www.getdrip.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://www.getdrip.com/oauth/token"
    ACCESS_TOKEN_METHOD = "POST"

    def get_user_id(self, details, response):
        return details["email"]

    def get_user_details(self, response):
        return {
            "email": response["users"][0]["email"],
            "fullname": response["users"][0]["name"],
            "username": response["users"][0]["email"],
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://api.getdrip.com/v2/user",
            headers={"Authorization": f"Bearer {access_token}"},
        )



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""

from urllib.parse import urljoin

from ..utils import append_slash
from .oauth import BaseOAuth2


class OktaMixin(BaseOAuth2):
    def api_url(self):
        return append_slash(self.setting("API_URL"))

    def authorization_url(self):
        return self._url("v1/authorize")

    def access_token_url(self):
        return self._url("v1/token")

    def _url(self, path):
        return urljoin(append_slash(self.setting("API_URL")), path)

    def oidc_config(self):
        return self.get_json(
            self._url(
                "/.well-known/openid-configuration?client_id={}".format(
                    self.setting("KEY")
                )
            )
        )


class OktaOAuth2(OktaMixin, BaseOAuth2):
    """Okta OAuth authentication backend"""

    name = "okta-oauth2"
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = " "
    ID_KEY = "preferred_username"

    DEFAULT_SCOPE = ["openid", "profile", "email"]
    EXTRA_DATA = [
        ("refresh_token", "refresh_token", True),
        ("expires_in", "expires"),
        ("token_type", "token_type", True),
    ]

    def get_user_details(self, response):
        """Return user details from Okta account"""
        return {
            "username": response.get("preferred_username"),
            "email": response.get("email") or "",
            "first_name": response.get("given_name"),
            "last_name": response.get("family_name"),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from Okta"""
        return self.get_json(
            self._url("v1/userinfo"),
            headers={
                "Authorization": f"Bearer {access_token}",
            },
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class LyftOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.lyft.LyftOAuth2"
    user_data_url = "https://api.lyft.com/v1/profile"
    access_token_body = json.dumps(
        {
            "access_token": "atoken_foo",
            "refresh_token": "rtoken_bar",
            "token_type": "bearer",
            "expires_in": 3600,
            "scope": "public profile rides.read rides.request",
            "id": "user_foobar",
        }
    )
    user_data_body = json.dumps({"id": "user_foobar"})
    expected_username = "user_foobar"

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import time

from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class ItembaseOAuth2(BaseOAuth2):
    name = "itembase"
    ID_KEY = "uuid"
    AUTHORIZATION_URL = "https://accounts.itembase.com/oauth/v2/auth"
    ACCESS_TOKEN_URL = "https://accounts.itembase.com/oauth/v2/token"
    USER_DETAILS_URL = "https://users.itembase.com/v1/me"
    ACTIVATION_ENDPOINT = "https://solutionservice.itembase.com/activate"
    DEFAULT_SCOPE = ["user.minimal"]
    EXTRA_DATA = [
        ("access_token", "access_token"),
        ("token_type", "token_type"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),  # seconds to expiration
        ("expires", "expires"),  # expiration timestamp in UTC
        ("uuid", "uuid"),
        ("username", "username"),
        ("email", "email"),
        ("first_name", "first_name"),
        ("middle_name", "middle_name"),
        ("last_name", "last_name"),
        ("name_format", "name_format"),
        ("locale", "locale"),
        ("preferred_currency", "preferred_currency"),
    ]

    def add_expires(self, data):
        data["expires"] = int(time.time()) + data.get("expires_in", 0)
        return data

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        data = BaseOAuth2.extra_data(
            self, user, uid, response, details=details, *args, **kwargs
        )
        return self.add_expires(data)

    def process_refresh_token_response(self, response, *args, **kwargs):
        data = BaseOAuth2.process_refresh_token_response(
            self, response, *args, **kwargs
        )
        return self.add_expires(data)

    def get_user_details(self, response):
        """Return user details from Itembase account"""
        return response

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            self.USER_DETAILS_URL, headers={"Authorization": f"Bearer {access_token}"}
        )

    def activation_data(self, response):
        # returns activation_data dict with activation_url inside
        # see http://developers.itembase.com/authentication/activation
        return self.get_json(
            self.ACTIVATION_ENDPOINT,
            headers={"Authorization": "Bearer {}".format(response["access_token"])},
        )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        state = self.validate_state()
        self.process_error(self.data)
        # itembase needs GET request with params instead of just data
        response = self.request_access_token(
            self.access_token_url(),
            params=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )


class ItembaseOAuth2Sandbox(ItembaseOAuth2):
    name = "itembase-sandbox"
    AUTHORIZATION_URL = "http://sandbox.accounts.itembase.io/oauth/v2/auth"
    ACCESS_TOKEN_URL = "http://sandbox.accounts.itembase.io/oauth/v2/token"
    USER_DETAILS_URL = "http://sandbox.users.itembase.io/v1/me"
    ACTIVATION_ENDPOINT = "http://sandbox.solutionservice.itembase.io/activate"



"""
Coinbase OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/coinbase.html
"""

from .oauth import BaseOAuth2


class CoinbaseOAuth2(BaseOAuth2):
    name = "coinbase"
    SCOPE_SEPARATOR = "+"
    DEFAULT_SCOPE = ["user", "balance"]
    AUTHORIZATION_URL = "https://www.coinbase.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.coinbase.com/oauth/token"
    REVOKE_TOKEN_URL = "https://api.coinbase.com/oauth/revoke"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response["data"]["id"]

    def get_user_details(self, response):
        """Return user details from Coinbase account"""
        user_data = response["data"]
        email = user_data.get("email", "")
        name = user_data["name"]
        username = user_data.get("username")
        fullname, first_name, last_name = self.get_user_names(name)
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.coinbase.com/v2/user",
            headers={"Authorization": "Bearer " + access_token},
        )



from __future__ import annotations

import base64
import hashlib
from typing import TYPE_CHECKING, Any
from urllib.parse import urlencode

from oauthlib.oauth1 import SIGNATURE_TYPE_AUTH_HEADER
from requests_oauthlib import OAuth1

from ..exceptions import (
    AuthCanceled,
    AuthException,
    AuthFailed,
    AuthMissingParameter,
    AuthStateForbidden,
    AuthStateMissing,
    AuthTokenError,
    AuthUnknownError,
)
from ..utils import (
    constant_time_compare,
    handle_http_errors,
    parse_qs,
    url_add_parameters,
)
from .base import BaseAuth

if TYPE_CHECKING:
    from collections.abc import Mapping, MutableMapping


class OAuthAuth(BaseAuth):
    """OAuth authentication backend base class.

    Settings will be inspected to get more values names that should be
    stored on extra_data field. The setting name is created following the
    pattern SOCIAL_AUTH_<uppercase current backend name>_EXTRA_DATA.

    access_token is always stored.

    URLs settings:
        AUTHORIZATION_URL       Authorization service url
        ACCESS_TOKEN_URL        Access token URL
    """

    AUTHORIZATION_URL = ""
    ACCESS_TOKEN_URL = ""
    ACCESS_TOKEN_METHOD = "GET"
    REVOKE_TOKEN_URL = None
    REVOKE_TOKEN_METHOD = "POST"
    ID_KEY = "id"
    SCOPE_PARAMETER_NAME = "scope"
    DEFAULT_SCOPE = None
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super().extra_data(user, uid, response, details, *args, **kwargs)
        data["access_token"] = response.get("access_token", "") or kwargs.get(
            "access_token"
        )
        return data

    def state_token(self):
        """Generate csrf token to include as state parameter."""
        return self.strategy.random_string(32)

    def get_or_create_state(self):
        if self.STATE_PARAMETER or self.REDIRECT_STATE:
            # Store state in session for further request validation. The state
            # value is passed as state parameter (as specified in OAuth2 spec),
            # but also added to redirect, that way we can still verify the
            # request if the provider doesn't implement the state parameter.
            # Reuse token if any.
            name = self.name + "_state"
            state = self.strategy.session_get(name)
            if state is None:
                state = self.state_token()
                self.strategy.session_set(name, state)
        else:
            state = None
        return state

    def get_session_state(self):
        return self.strategy.session_get(self.name + "_state")

    def get_request_state(self):
        request_state = self.data.get("state") or self.data.get("redirect_state")
        if request_state and isinstance(request_state, list):
            request_state = request_state[0]
        return request_state

    def validate_state(self):
        """Validate state value. Raises exception on error, returns state
        value if valid."""
        if not self.STATE_PARAMETER and not self.REDIRECT_STATE:
            return None
        state = self.get_session_state()
        request_state = self.get_request_state()
        if not request_state:
            raise AuthMissingParameter(self, "state")
        if not state:
            raise AuthStateMissing(self, "state")
        if not constant_time_compare(request_state, state):
            raise AuthStateForbidden(self)
        return state

    def get_redirect_uri(self, state=None):
        """Build redirect with redirect_state parameter."""
        uri = self.redirect_uri
        if self.REDIRECT_STATE and state:
            uri = url_add_parameters(uri, {"redirect_state": state})
        return uri

    def get_scope(self):
        """Return list with needed access scope"""
        scope = self.setting("SCOPE", [])
        if not self.setting("IGNORE_DEFAULT_SCOPE", False):
            scope = scope + (self.DEFAULT_SCOPE or [])
        return scope

    def get_scope_argument(self):
        param = {}
        scope = self.get_scope()
        if scope:
            param[self.SCOPE_PARAMETER_NAME] = self.SCOPE_SEPARATOR.join(scope)
        return param

    def user_data(self, access_token, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service. Implement in subclass"""
        return {}

    def authorization_url(self) -> str:
        return self.AUTHORIZATION_URL

    def access_token_url(self) -> str:
        return self.ACCESS_TOKEN_URL

    def revoke_token_url(self, token, uid) -> str | None:
        return self.REVOKE_TOKEN_URL

    def revoke_token_params(self, token, uid) -> dict[str, Any]:
        return {}

    def revoke_token_headers(self, token, uid) -> dict[str, Any]:
        return {}

    def process_revoke_token_response(self, response):
        return response.status_code == 200

    def revoke_token(self, token, uid):
        if self.REVOKE_TOKEN_URL:
            url = self.revoke_token_url(token, uid)
            params = self.revoke_token_params(token, uid)
            headers = self.revoke_token_headers(token, uid)
            data = urlencode(params) if self.REVOKE_TOKEN_METHOD != "GET" else None
            response = self.request(
                url,
                params=params,
                headers=headers,
                data=data,
                method=self.REVOKE_TOKEN_METHOD,
            )
            return self.process_revoke_token_response(response)
        return None


class BaseOAuth1(OAuthAuth):
    """Consumer based mechanism OAuth authentication, fill the needed
    parameters to communicate properly with authentication service.

    URLs settings:
        REQUEST_TOKEN_URL       Request token URL

    """

    REQUEST_TOKEN_URL = ""
    REQUEST_TOKEN_METHOD = "GET"
    OAUTH_TOKEN_PARAMETER_NAME = "oauth_token"
    REDIRECT_URI_PARAMETER_NAME = "redirect_uri"
    UNATHORIZED_TOKEN_SUFIX = "unauthorized_token_name"

    def auth_url(self) -> str | bytes | None:
        """Return redirect url"""
        token = self.set_unauthorized_token()
        return self.oauth_authorization_request(token)

    def process_error(self, data):
        if "oauth_problem" in data:
            if data["oauth_problem"] == "user_refused":
                raise AuthCanceled(self, "User refused the access")
            raise AuthUnknownError(self, "Error was " + data["oauth_problem"])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Return user, might be logged in"""
        # Multiple unauthorized tokens are supported (see #521)
        self.process_error(self.data)
        self.validate_state()
        token = self.get_unauthorized_token()
        access_token = self.access_token(token)
        return self.do_auth(access_token, *args, **kwargs)

    @handle_http_errors
    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        if not isinstance(access_token, dict):
            access_token = parse_qs(access_token)
        data = self.user_data(access_token)
        if data is not None and "access_token" not in data:
            data["access_token"] = access_token
        kwargs.update({"response": data, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_unauthorized_token(self):
        name = self.name + self.UNATHORIZED_TOKEN_SUFIX
        unauthed_tokens = self.strategy.session_get(name, [])
        if not unauthed_tokens:
            raise AuthTokenError(self, "Missing unauthorized token")

        data_token = self.data.get(self.OAUTH_TOKEN_PARAMETER_NAME)

        if data_token is None:
            raise AuthTokenError(self, "Missing unauthorized token")

        token = None
        for utoken in unauthed_tokens:
            orig_utoken = utoken
            if not isinstance(utoken, dict):
                utoken = parse_qs(utoken)
            if utoken.get(self.OAUTH_TOKEN_PARAMETER_NAME) == data_token:
                self.strategy.session_set(
                    name, list(set(unauthed_tokens) - {orig_utoken})
                )
                token = utoken
                break
        else:
            raise AuthTokenError(self, "Incorrect tokens")
        return token

    def set_unauthorized_token(self):
        token = self.unauthorized_token()
        name = self.name + self.UNATHORIZED_TOKEN_SUFIX
        tokens = [*self.strategy.session_get(name, []), token]
        self.strategy.session_set(name, tokens)
        return token

    def request_token_extra_arguments(self):
        """Return extra arguments needed on request-token process"""
        return self.setting("REQUEST_TOKEN_EXTRA_ARGUMENTS", {})

    def unauthorized_token(self):
        """Return request for unauthorized token (first stage)"""
        params = self.request_token_extra_arguments()
        params.update(self.get_scope_argument())
        key, secret = self.get_key_and_secret()
        state = self.get_or_create_state()
        response = self.request(
            self.REQUEST_TOKEN_URL,
            params=params,
            auth=OAuth1(key, secret, callback_uri=self.get_redirect_uri(state)),
            method=self.REQUEST_TOKEN_METHOD,
        )
        content = response.content
        if response.encoding or response.apparent_encoding:
            content = content.decode(response.encoding or response.apparent_encoding)
        else:
            content = response.content.decode()
        return content

    def oauth_authorization_request(self, token):
        """Generate OAuth request to authorize token."""
        if not isinstance(token, dict):
            token = parse_qs(token)
        params = self.auth_extra_arguments() or {}
        params.update(self.get_scope_argument())
        params[self.OAUTH_TOKEN_PARAMETER_NAME] = token.get(
            self.OAUTH_TOKEN_PARAMETER_NAME
        )
        state = self.get_or_create_state()
        params[self.REDIRECT_URI_PARAMETER_NAME] = self.get_redirect_uri(state)
        return url_add_parameters(self.authorization_url(), params)

    def oauth_auth(
        self, token=None, oauth_verifier=None, signature_type=SIGNATURE_TYPE_AUTH_HEADER
    ):
        key, secret = self.get_key_and_secret()
        oauth_verifier = oauth_verifier or self.data.get("oauth_verifier")
        if token:
            resource_owner_key = token.get("oauth_token")
            resource_owner_secret = token.get("oauth_token_secret")
            if not resource_owner_key:
                raise AuthTokenError(self, "Missing oauth_token")
            if not resource_owner_secret:
                raise AuthTokenError(self, "Missing oauth_token_secret")
        else:
            resource_owner_key = None
            resource_owner_secret = None
        state = self.get_or_create_state()
        return OAuth1(
            key,
            secret,
            resource_owner_key=resource_owner_key,
            resource_owner_secret=resource_owner_secret,
            callback_uri=self.get_redirect_uri(state),
            verifier=oauth_verifier,
            signature_type=signature_type,
        )

    def oauth_request(self, token, url, params=None, method="GET"):
        """Generate OAuth request, setups callback url"""
        return self.request(
            url, method=method, params=params, auth=self.oauth_auth(token)
        )

    def access_token(self, token):
        """Return request for access token value"""
        return self.get_querystring(
            self.access_token_url(),
            auth=self.oauth_auth(token),
            method=self.ACCESS_TOKEN_METHOD,
        )


class BaseOAuth2(OAuthAuth):
    """Base class for OAuth2 providers.

    OAuth2 details at:
        https://datatracker.ietf.org/doc/html/rfc6749
    """

    REFRESH_TOKEN_URL = None
    REFRESH_TOKEN_METHOD = "POST"
    RESPONSE_TYPE = "code"
    REDIRECT_STATE = True
    STATE_PARAMETER = True
    USE_BASIC_AUTH = False

    def use_basic_auth(self) -> bool:
        return self.USE_BASIC_AUTH

    def auth_params(self, state=None) -> MutableMapping[str, Any]:
        client_id, client_secret = self.get_key_and_secret()
        params = {"client_id": client_id, "redirect_uri": self.get_redirect_uri(state)}
        if self.STATE_PARAMETER and state:
            params["state"] = state
        if self.RESPONSE_TYPE:
            params["response_type"] = self.RESPONSE_TYPE
        return params

    def auth_url(self) -> str | bytes | None:
        """Return redirect url"""
        state = self.get_or_create_state()
        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())

        # when self.REDIRECT_STATE is False, redirect_uri matching is strictly enforced,
        # so match the providers value exactly.
        return url_add_parameters(
            self.authorization_url(), params, not self.REDIRECT_STATE
        )

    def auth_complete_params(self, state=None):
        params = {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "redirect_uri": self.get_redirect_uri(state),
        }
        if not self.use_basic_auth():
            client_id, client_secret = self.get_key_and_secret()
            params.update(
                {
                    "client_id": client_id,
                    "client_secret": client_secret,
                }
            )
        return params

    def auth_complete_credentials(self):
        if self.use_basic_auth():
            return self.get_key_and_secret()
        return None

    def auth_headers(self) -> Mapping[str, str | bytes]:
        return {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token, token_type, and extra defined names to store in
        extra_data field"""
        data = super().extra_data(user, uid, response, details=details, *args, **kwargs)
        data["token_type"] = response.get("token_type") or kwargs.get("token_type")
        return data

    def request_access_token(self, *args, **kwargs):
        return self.get_json(*args, **kwargs)

    def process_error(self, data):
        if data.get("error"):
            if "denied" in data["error"] or "cancelled" in data["error"]:
                raise AuthCanceled(self, data.get("error_description", ""))
            raise AuthFailed(self, data.get("error_description") or data["error"])
        if "denied" in data:
            raise AuthCanceled(self, data["denied"])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()
        data, params = None, None
        if self.ACCESS_TOKEN_METHOD == "GET":
            params = self.auth_complete_params(state)
        else:
            data = self.auth_complete_params(state)

        response = self.request_access_token(
            self.access_token_url(),
            data=data,
            params=params,
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )

    @handle_http_errors
    def do_auth(self, access_token, *args, **kwargs):
        """Finish the auth process once the access_token was retrieved"""
        data = self.user_data(access_token, *args, **kwargs)
        response = kwargs.get("response") or {}
        response.update(data or {})
        if "access_token" not in response:
            response["access_token"] = access_token
        kwargs.update({"response": response, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def refresh_token_params(self, token, *args, **kwargs):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "refresh_token": token,
            "grant_type": "refresh_token",
            "client_id": client_id,
            "client_secret": client_secret,
        }

    def process_refresh_token_response(self, response, *args, **kwargs):
        return response.json()

    def refresh_token(self, token, *args, **kwargs):
        params = self.refresh_token_params(token, *args, **kwargs)
        url = self.refresh_token_url()
        method = self.REFRESH_TOKEN_METHOD
        key = "params" if method == "GET" else "data"
        request_args = {"headers": self.auth_headers(), "method": method, key: params}
        request = self.request(url, **request_args)
        return self.process_refresh_token_response(request, *args, **kwargs)

    def refresh_token_url(self):
        return self.REFRESH_TOKEN_URL or self.access_token_url()


class BaseOAuth2PKCE(BaseOAuth2):
    """
    Base class for providers using OAuth2 with Proof Key for Code Exchange (PKCE).

    OAuth2 details at:
        https://datatracker.ietf.org/doc/html/rfc6749
    PKCE details at:
        https://datatracker.ietf.org/doc/html/rfc7636
    """

    PKCE_DEFAULT_CODE_CHALLENGE_METHOD = "s256"
    PKCE_DEFAULT_CODE_VERIFIER_LENGTH = 32
    DEFAULT_USE_PKCE = True

    def create_code_verifier(self):
        name = f"{self.name}_code_verifier"
        code_verifier_len = self.setting(
            "PKCE_CODE_VERIFIER_LENGTH", default=self.PKCE_DEFAULT_CODE_VERIFIER_LENGTH
        )
        code_verifier = self.strategy.random_string(code_verifier_len)
        self.strategy.session_set(name, code_verifier)
        return code_verifier

    def get_code_verifier(self):
        name = f"{self.name}_code_verifier"
        return self.strategy.session_get(name)

    def generate_code_challenge(self, code_verifier, challenge_method):
        method = challenge_method.lower()
        if method == "s256":
            hashed = hashlib.sha256(code_verifier.encode()).digest()
            encoded = base64.urlsafe_b64encode(hashed)
            return encoded.decode().replace("=", "")  # remove padding
        if method == "plain":
            return code_verifier
        raise AuthException("Unsupported code challenge method.")

    def auth_params(self, state=None):
        params = super().auth_params(state=state)

        if self.setting("USE_PKCE", default=self.DEFAULT_USE_PKCE):
            code_challenge_method = self.setting(
                "PKCE_CODE_CHALLENGE_METHOD",
                default=self.PKCE_DEFAULT_CODE_CHALLENGE_METHOD,
            )
            code_verifier = self.create_code_verifier()
            code_challenge = self.generate_code_challenge(
                code_verifier, code_challenge_method
            )
            params["code_challenge_method"] = code_challenge_method
            params["code_challenge"] = code_challenge
        return params

    def auth_complete_params(self, state=None):
        params = super().auth_complete_params(state=state)

        if self.setting("USE_PKCE", default=self.DEFAULT_USE_PKCE):
            code_verifier = self.get_code_verifier()
            params["code_verifier"] = code_verifier

        return params



from uuid import uuid4

from ..utils import module_member, slugify

USER_FIELDS = ["username", "email"]


def get_username(strategy, details, backend, user=None, *args, **kwargs):
    if "username" not in backend.setting("USER_FIELDS", USER_FIELDS):
        return None
    storage = strategy.storage

    if not user:
        email_as_username = backend.setting("USERNAME_IS_FULL_EMAIL", False)
        uuid_length = backend.setting("UUID_LENGTH", 16)
        max_length = storage.user.username_max_length()
        do_slugify = backend.setting("SLUGIFY_USERNAMES", False)
        do_clean = backend.setting("CLEAN_USERNAMES", True)

        def identity_func(val):
            return val

        if do_clean:
            override_clean = backend.setting("CLEAN_USERNAME_FUNCTION")
            if override_clean:
                clean_func = module_member(override_clean)
            else:
                clean_func = storage.user.clean_username
        else:
            clean_func = identity_func

        if do_slugify:
            override_slug = backend.setting("SLUGIFY_FUNCTION")
            slug_func = module_member(override_slug) if override_slug else slugify
        else:
            slug_func = identity_func

        if email_as_username and details.get("email"):
            username = details["email"]
        elif details.get("username"):
            username = details["username"]
        else:
            username = uuid4().hex

        short_username = (
            username[: max_length - uuid_length] if max_length is not None else username
        )
        final_username = slug_func(clean_func(username[:max_length]))

        # Generate a unique username for current user using username
        # as base but adding a unique hash at the end. Original
        # username is cut to avoid any field max_length.
        # The final_username may be empty and will skip the loop.
        while not final_username or storage.user.user_exists(username=final_username):
            username = short_username + uuid4().hex[:uuid_length]
            final_username = slug_func(clean_func(username[:max_length]))
    else:
        final_username = storage.user.get_username(user)
    return {"username": final_username}


def create_user(strategy, details, backend, user=None, *args, **kwargs):
    if user:
        return {"is_new": False}

    fields = {
        name: kwargs.get(name, details.get(name))
        for name in backend.setting("USER_FIELDS", USER_FIELDS)
    }
    if not fields:
        return None

    # Allow overriding the email field if desired by application specification
    if backend.setting("FORCE_EMAIL_LOWERCASE", False):
        emailfield = fields.get("email")
        if emailfield:
            fields["email"] = emailfield.lower()

    return {"is_new": True, "user": strategy.create_user(**fields)}


def user_details(strategy, details, backend, user=None, *args, **kwargs):
    """Update user details using data from provider."""
    if not user:
        return

    changed = False  # flag to track changes

    # Default protected user fields (username, id, pk and email) can be ignored
    # by setting the SOCIAL_AUTH_NO_DEFAULT_PROTECTED_USER_FIELDS to True
    if strategy.setting("NO_DEFAULT_PROTECTED_USER_FIELDS", backend=backend) is True:
        protected = ()
    else:
        protected = (
            "username",
            "id",
            "pk",
            "email",
            "password",
            "is_active",
            "is_staff",
            "is_superuser",
        )

    protected = protected + tuple(
        strategy.setting("PROTECTED_USER_FIELDS", [], backend=backend)
    )

    # Update user model attributes with the new data sent by the current
    # provider. Update on some attributes is disabled by default, for
    # example username and id fields. It's also possible to disable update
    # on fields defined in SOCIAL_AUTH_PROTECTED_USER_FIELDS.
    field_mapping = strategy.setting("USER_FIELD_MAPPING", {}, backend=backend)
    for name, value in details.items():
        # Convert to existing user field if mapping exists
        name = field_mapping.get(name, name)
        if value is None or not hasattr(user, name) or name in protected:
            continue

        current_value = getattr(user, name, None)
        if current_value == value:
            continue

        immutable_fields = tuple(
            strategy.setting("IMMUTABLE_USER_FIELDS", [], backend=backend)
        )
        if name in immutable_fields and current_value:
            continue

        changed = True
        setattr(user, name, value)

    if changed:
        strategy.storage.user.changed(user)



"""
Facebook Limited Login backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/facebook.html
"""

from ..exceptions import AuthTokenError
from .open_id_connect import OpenIdConnectAuth


class FacebookLimitedLogin(OpenIdConnectAuth):
    """Facebook Limited Login (OIDC) backend"""

    name = "facebook-limited-login"
    OIDC_ENDPOINT = "https://www.facebook.com"
    ACCESS_TOKEN_URL = "https://facebook.com/dialog/oauth/"
    ID_TOKEN_MAX_AGE = 3600

    def authenticate(self, *args, **kwargs):
        if (
            "backend" not in kwargs
            or kwargs["backend"].name != self.name
            or "strategy" not in kwargs
            or "response" not in kwargs
        ):
            return None

        # Replace response with the decoded JWT
        raw_jwt = kwargs.get("response", {}).get("access_token")
        kwargs["response"] = self.validate_and_return_id_token(raw_jwt, "")
        return super().authenticate(*args, **kwargs)

    def get_user_details(self, response):
        return {
            "fullname": response.get("name"),
            "email": response.get("email"),
            "picture": response.get("picture"),
        }

    def user_data(self, access_token, *args, **kwargs):
        # We don't have an access token to call any API for the user details.
        return None

    def validate_claims(self, id_token):
        try:
            super().validate_claims(id_token)
        except AuthTokenError as e:
            if "Incorrect id_token: nonce" in e.args:
                # Ignore errors about nonce. We can't validate it since it's not generated server-side.
                return
            raise



"""
Skyrock OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/skyrock.html
"""

from .oauth import BaseOAuth1


class SkyrockOAuth(BaseOAuth1):
    """Skyrock OAuth authentication backend"""

    name = "skyrock"
    ID_KEY = "id_user"
    AUTHORIZATION_URL = "https://api.skyrock.com/v2/oauth/authenticate"
    REQUEST_TOKEN_URL = "https://api.skyrock.com/v2/oauth/initiate"
    ACCESS_TOKEN_URL = "https://api.skyrock.com/v2/oauth/token"
    EXTRA_DATA = [("id", "id")]

    def get_user_details(self, response):
        """Return user details from Skyrock account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response["firstname"], last_name=response["name"]
        )
        return {
            "username": response["username"],
            "email": response["email"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token):
        """Return user data provided"""
        return self.get_json(
            "https://api.skyrock.com/v2/user/get.json",
            auth=self.oauth_auth(access_token),
        )



import json
from urllib.parse import urlencode

from httpretty import HTTPretty

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class DiscsogsOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    _test_token = "lalala123boink"
    backend_path = "social_core.backends.discogs.DiscogsOAuth1"
    expected_username = "rodneyfool"
    raw_complete_url = (
        f"/complete/{0}/?oauth_verifier=wimblewomblefartfart&oauth_token={_test_token}"
    )

    access_token_body = json.dumps(
        {"access_token": _test_token, "token_type": "bearer"}
    )
    request_token_body = urlencode(
        {
            "oauth_token": _test_token,
            "oauth_token_secret": "xyz789",
            "oauth_callback_confirmed": "true",
        }
    )

    user_data_body = json.dumps(
        {
            "profile": "I am a software developer for Discogs.\r\n\r\n[img=http://i.imgur.com/IAk3Ukk.gif]",
            "wantlist_url": "https://api.discogs.com/users/rodneyfool/wants",
            "rank": 149,
            "num_pending": 61,
            "id": 1578108,
            "num_for_sale": 0,
            "home_page": "",
            "location": "I live in the good ol' Pacific NW",
            "collection_folders_url": "https://api.discogs.com/users/rodneyfool/collection/folders",
            "username": expected_username,
            "collection_fields_url": "https://api.discogs.com/users/rodneyfool/collection/fields",
            "releases_contributed": 5,
            "registered": "2012-08-15T21:13:36-07:00",
            "rating_avg": 3.47,
            "num_collection": 78,
            "releases_rated": 116,
            "num_lists": 0,
            "name": "Rodney",
            "num_wantlist": 160,
            "inventory_url": "https://api.discogs.com/users/rodneyfool/inventory",
            "avatar_url": "http://www.gravatar.com/avatar/55502f40dc8b7c769880b10874abc9d0?s=52&r=pg&d=mm",
            "banner_url": (
                "https://img.discogs.com/dhuJe-pRJmod7hN3cdVi2PugEh4=/1600x400/"
                "filters:strip_icc():format(jpeg)/discogs-banners/B-1578108-user-1436314164-9231.jpg.jpg"
            ),
            "uri": "https://www.discogs.com/user/rodneyfool",
            "resource_url": "https://api.discogs.com/users/rodneyfool",
            "buyer_rating": 100.00,
            "buyer_rating_stars": 5,
            "buyer_num_ratings": 144,
            "seller_rating": 100.00,
            "seller_rating_stars": 5,
            "seller_num_ratings": 21,
            "curr_abbr": "USD",
        }
    )

    def _mock(self):
        HTTPretty.register_uri(
            HTTPretty.GET,
            uri="https://api.discogs.com/oauth/identity",
            status=200,
            body=json.dumps(
                {
                    "id": 1,
                    "username": self.expected_username,
                    "resource_url": f"https://api.discogs.com/users/{self.expected_username}",
                    "consumer_name": "SocialCore Discogs Test",
                }
            ),
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            f"https://api.discogs.com/users/{self.expected_username}",
            status=200,
            body=self.user_data_body,
        )

    def test_login(self):
        self._mock()
        self.do_login()

    def test_partial_pipeline(self):
        self._mock()
        self.do_partial_pipeline()



import hashlib
import hmac
import time

from ..exceptions import AuthFailed, AuthMissingParameter
from ..utils import handle_http_errors
from .base import BaseAuth


class TelegramAuth(BaseAuth):
    name = "telegram"
    ID_KEY = "id"

    def verify_data(self, response):
        bot_token = self.setting("BOT_TOKEN")
        if bot_token is None:
            raise AuthMissingParameter("telegram", "SOCIAL_AUTH_TELEGRAM_BOT_TOKEN")

        received_hash_string = response.get("hash")
        auth_date = response.get("auth_date")

        if received_hash_string is None or auth_date is None:
            raise AuthMissingParameter("telegram", "hash or auth_date")

        data_check_string = [f"{k}={v}" for k, v in response.items() if k != "hash"]
        data_check_string = "\n".join(sorted(data_check_string))
        secret_key = hashlib.sha256(bot_token.encode()).digest()
        built_hash = hmac.new(
            secret_key, msg=data_check_string.encode(), digestmod=hashlib.sha256
        ).hexdigest()
        current_timestamp = int(time.time())
        auth_timestamp = int(auth_date)
        if current_timestamp - auth_timestamp > 86400:
            raise AuthFailed("telegram", "Auth date is outdated")
        if built_hash != received_hash_string:
            raise AuthFailed("telegram", "Invalid hash supplied")

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        return response

    def get_user_details(self, response):
        first_name = response.get("first_name", "")
        last_name = response.get("last_name", "")
        fullname = f"{first_name} {last_name}".strip()
        return {
            "username": response.get("username") or str(response[self.ID_KEY]),
            "first_name": first_name,
            "last_name": last_name,
            "fullname": fullname,
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        response = self.data
        self.verify_data(response)
        kwargs.update({"response": self.data, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)



from ..exceptions import NotAllowedToDisconnect


def allowed_to_disconnect(
    strategy, user, name, user_storage, association_id=None, *args, **kwargs
):
    if not user_storage.allowed_to_disconnect(user, name, association_id):
        raise NotAllowedToDisconnect


def get_entries(
    strategy, user, name, user_storage, association_id=None, *args, **kwargs
):
    return {
        "entries": user_storage.get_social_auth_for_user(user, name, association_id)
    }


def revoke_tokens(strategy, entries, *args, **kwargs):
    revoke_tokens = strategy.setting("REVOKE_TOKENS_ON_DISCONNECT", False)
    if revoke_tokens:
        for entry in entries:
            if "access_token" in entry.extra_data:
                backend = entry.get_backend_instance(strategy)
                backend.revoke_token(entry.extra_data["access_token"], entry.uid)


def disconnect(strategy, entries, user_storage, *args, **kwargs):
    for entry in entries:
        user_storage.disconnect(entry)



"""
SimpleLogin OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/simplelogin.html
"""

from .oauth import BaseOAuth2


class SimpleLoginOAuth2(BaseOAuth2):
    """SimpleLogin OAuth authentication backend"""

    name = "simplelogin"
    AUTHORIZATION_URL = "https://app.simplelogin.io/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://app.simplelogin.io/oauth2/token"
    ACCESS_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ("name", "name"),
        ("email", "email"),
        ("avatar_url", "avatar_url"),
    ]

    # endpoint to get user info
    USERINFO_URL = "https://app.simplelogin.io/oauth2/userinfo"

    def get_user_details(self, response):
        """Return user details from SimpleLogin account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": response.get("email"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(self.USERINFO_URL, params={"access_token": access_token})
